#!/usr/bin/python

'''
   File name: RulesLib.py
   Author: Wayne Parham
   Date created: 03/01/2018
'''


import base64, fnmatch, json, os, shlex, string, subprocess, sys, time
try:
   from configparser import ConfigParser
except ImportError:
   from ConfigParser import ConfigParser


POLLING_INTERVAL        =  0.1          # 1/10th second
EXCESSIVE_DELAY         =  600          # one minute
EXCESSIVE_TIMEOUT       = 3600          # one hour
LINE_BUFSIZE            = 1024          # line bufffer size

quote                   = '"'
configFile              = "RulesNgn.ini"
manifest_file           = "manifest"
rules_identifier        = ".csv"
datadir                 = "."
logdir                  = "."
cleanup_enabled         = False
attachments_enabled     = True
send_to_jira            = True

user                    = ""
password                = ""
endpoint                = ""
org                     = ""
project                 = ""
repo                    = "" 
entrypoint              = ""
cURLcmd                 = "curl -u "+user+":"+password+" "+endpoint+" -v -F org="+org+" -F project="+project+" -F repo="+repo+" -F entryPoint="+entrypoint+" -F request="

if( os.path.isfile(configFile) ):
   print( "==> Config file found, loading ["+configFile+"]" )
   try:
      config = ConfigParser()
   except TypeError:
      config = ConfigParser.ConfigParser()
   config.read( configFile )
   try:
      datadir             = config.get( 'RulesEngine', 'datadir' )
   except:
      pass
   try:
      logdir              = config.get( 'RulesEngine', 'logdir' )
   except:
      pass
   try:
      rules_identifier    = config.get( 'RulesEngine', 'rules_identifier' )
   except:
      pass
   try:
      manifest_file       = config.get( 'ReportingInterface', 'manifest_file' )
   except:
      pass
   try:
      cleanup_enabled     = bit(config.get('ReportingInterface', 'cleanup_enabled'))
   except:
      pass
   try:
      attachments_enabled = bit(config.get('ReportingInterface', 'attachments_enabled'))
   except:
      pass
   try:
      send_to_jira        = bit(config.get('ReportingInterface', 'send_to_jira'))
   except:
      pass
   try:
      user                = config.get( 'ReportingInterface', 'user' )
   except:
      pass
   try:
      password            = config.get( 'ReportingInterface', 'password' )
   except:
      pass
   try:
      endpoint            = config.get( 'ReportingInterface', 'endpoint' )
   except:
      pass
   try:
      org                 = config.get( 'ReportingInterface', 'org' )
   except:
      pass
   try:
      project             = config.get( 'ReportingInterface', 'project' )
   except:
      pass
   try:
      repo                = config.get( 'ReportingInterface', 'repo' )
   except:
      pass
   try:
      entrypoint          = config.get( 'ReportingInterface', 'entrypoint' )
   except:
      pass
   try:
      POLLING_INTERVAL    = config.get( 'RulesEngine', 'polling_interval' )
   except:
      pass
   try:
      EXCESSIVE_DELAY     = config.get( 'RulesEngine', 'excessive_delay' )
   except:
      pass
   try:
      EXCESSIVE_TIMEOUT   = config.get( 'RulesEngine', 'excessive_timeout' )
   except:
      pass
   try:
      LINE_BUFSIZE        = config.get( 'RulesEngine', 'line_bufsize' )
   except:
      pass

   if( ('rulesngn' in sys.argv[0].lower()) or ('testx' in sys.argv[0].lower()) ):
      print( "==> Data dir:         "+datadir )
      print( "==> Log  dir:         "+logdir )
   if( 'testx' in sys.argv[0].lower() ):
      print( "==> Attachments:      "+str(attachments_enabled) )
      print( "==> Cleanup:          "+str(cleanup_enabled) )
      print( "==> Jira:             "+str(send_to_jira) )
else:
   print( "==> No "+configFile+" configuration file found.  Using defaults." )


def bit( string ):
   if( (string.lower() == "true") or (string.lower() == "yes") or (string.lower() == "y") ):
      bit = True
   else:
      bit = False
   return( bit )


def IsListofLists( unknown_list ):
   return( any(isinstance(l, list) for l in unknown_list) )


def getDateTime():
   now = time.localtime()
   localtime = time.strftime( '%Y-%m-%d %H:%M:%S', now )
   return( localtime )


def encode( object ):
   try:
      retstring = base64.b64encode( object.encode() )
   except:
      retstring = base64.b64encode( object )
   return( retstring )


def decode( string ):
   return( base64.b64decode(string) )


def WriteFile( filename, data ):
   retval = 0
   try:
      targetfile = open( filename, "w" )
      targetfile.write( data )
      targetfile.close
   except IOError:
      retval = -1
      print( "==> ERROR: [" + filename + "] can't be opened for writing." )
   return( retval )


def WriteEncodedFile( filename, data ):
   retval = 0
   try:
      exhibitfile = open( filename, "wb" )
      decoded_data = base64.b64decode( data )
      ByteArray = bytearray( decoded_data )
      exhibitfile.write( ByteArray )
      exhibitfile.close
   except IOError:
      retval = -1
      print( "==> ERROR: [" + filename + "] can't be opened for writing." )
   return( retval )


def EncodeFile( filename ):
   retval = -1
   try:
      datafile = open( filename, "rb" )
      data = datafile.read()
      retval = base64.b64encode(data).decode()
   except OSError:
      print( "==> ERROR: couldn't open [" + filename + "] for reading." )
   except IOError:
      print( "==> ERROR: No such file: [" + filename + "]." )
   return( retval )


def TouchFile( filename, times=None ):
   try:
      with open(filename, "a"):
         os.utime(filename, times)
      value = 0
   except IOError:
      value = -1
   return( value )


def WriteLine( filename, line, logfile ):
   try:
      targetfile = open( filename, "a" )
      targetfile.write( line )
      targetfile.close
   except IOError:
      info = "==> ERROR: [" + filename + "] can't be opened for writing."
      InfoLine( logfile, info )
   return


def InfoLine( filename, line ):
   print( line )
   line += "\n"
   try:
      targetfile = open( filename, "a" )
      targetfile.write( line )
      targetfile.close
   except IOError:
      print ("==> ERROR: [" + filename + "] can't be opened for writing.")
   sys.stdout.flush()
   return


def find( root_directory, search_pattern ):
   files = []
   search_pattern = search_pattern.lower()
   for dirName, subdirName, fileList in os.walk( root_directory ):
      for filename in fileList:
         if fnmatch.fnmatch( filename, search_pattern ):
            files.append( os.path.join(dirName, filename) ) 
   return( files )


def SearchFile( filename, search_phrase, logfile ):
   value = 1
   try:
      with open( filename,"r" ) as searchfile:
         line = searchfile.read()
         found = line.find( search_phrase )
         if( found != -1 ):
            value = 0
      searchfile.close
   except IOError:
      value = -1
      info = "==> ERROR: [" + filename + "] can't be opened for reading."
      InfoLine( logfile, info )
   return( value )


def DetermineFileType( filename ):
   if( filename.endswith(".bmp") ):
      filetype = "image/bmp"
   elif( filename.endswith(".csv") ):
      filetype = "text/csv"
   elif( filename.endswith(".doc") ):
      filetype = "application/msword"
   elif( filename.endswith(".gif") ):
      filetype = "image/gif"
   elif( filename.endswith(".gz") ):
      filetype = "application/x-gzip"
   elif( filename.endswith(".jpg") ):
      filetype = "image/jpeg"
   elif( filename.endswith(".log") ):
      filetype = "text/plain"
   elif( filename.endswith(".pdf") ):
      filetype = "application/pdf"
   elif( filename.endswith(".png") ):
      filetype = "image/png"
   elif( filename.endswith(".pl") ):
      filetype = "text/x-script.perl"
   elif( filename.endswith(".py") ):
      filetype = "application/x-python-code"
   elif( filename.endswith(".ppt") ):
      filetype = "application/mspowerpoint"
   elif( filename.endswith(".tar") ):
      filetype = "application/x-tar"
   elif( filename.endswith(".txt") ):
      filetype = "text/plain"
   elif( filename.endswith(".zip") ):
      filetype = "application/zip"
   else:
      filetype = "application/unknown"
   return( filetype )


def GetFilename( type ):
   milli = int(round(time.time() * 1000)) % 10000
   filename = type + str(milli).zfill(4) + ".dat"
   return( filename )


def runRules( rulesfile ):
   if ( rulesfile.endswith(rules_identifier) ):
      logfile = logdir  + "/" + rulesfile.rsplit("/")[-1].rsplit(".csv", 1)[0] + ".log"
      print( "==> INFO: Output is being logged in [" + logfile + "]." )

      if os.path.isfile(logfile):
         os.remove(logfile)

      ruleset = ReadRulesFile( rulesfile, logfile )

      if ruleset:
         return_value = ProcessRules( ruleset, logfile )

         localtime = getDateTime()

         if( return_value == 0 ):
            info = "==> PASS [" + rulesfile + "] - All tests succeeded at " + localtime 
            InfoLine( logfile, info )
         elif( return_value == -1 ):
            info = "==> EXIT [" + rulesfile + "] - Test execution stopped prior to completion at " + localtime
            InfoLine( logfile, info )
         else:
            info = "==> FAIL [" + rulesfile + "] - One or more test conditions failed at " + localtime
            InfoLine( logfile, info )
   else:
      print( "==> ERROR: " + rulesfile + " is not a ruleset file." )


def ReadRulesFile( rulesfile, logfile ):
   ruleset = []
   linecount = 0
   rulecount = 0
   info = "==> Processing [" + rulesfile + "] Rules File."
   InfoLine( logfile, info )
   try:
      rules = open( rulesfile, "r" )
      for line in rules:
         line = line.rstrip()
         if (len(line) > 0):
            if (line[0] != "0"):
               if (line[0] != "#"):
                  conditions, actions = ParseLine( line )
                  ruleset.append( conditions )
                  ruleset.append( actions )
                  rulecount += 1
         linecount += 1
      rules.close
      info = "==> Rules File [" + rulesfile + "] contains (" + str(linecount) + ") lines, (" + str(rulecount) + ") containing rules."
      InfoLine( logfile, info )
      return( ruleset )
   except IOError:
      info = "==> ERROR: [" + rulesfile + "] rules file can't be opened for reading."
      InfoLine( logfile, info )
   return


def ProcessRules( ruleset, logfile ):
   return_value = 0
   conditions, actions = BifurcateRules( ruleset )
   lastLine = len( conditions )
   localtime = getDateTime()
   info = "==> Start run at " + localtime
   InfoLine( logfile, info )
   for i in range( 0, lastLine ):
      if ( return_value == -1 ):
         break
      condition, action = IndividuateLines( i, conditions, actions )
      lastCommand = len( condition )
      lastTest = len( action )
      for c in range( 0, lastCommand ):
         command = IndividuateFields( c, condition )
         if ( len(command) > 0 ):
            if ( i < 9 ):
               info = "  " + str(i+1) + " cmd: " + command
               InfoLine( logfile, info )
            elif ( i < 99 ):
               info = " " + str(i+1) + " cmd: " + command
               InfoLine( logfile, info )
            else:
               info = str(i+1) + " cmd: " + command
               InfoLine( logfile, info )
         retcode, retoutput, reterr = PerformAction( command, logfile )
         if ( retcode == -1 ):
            return_value = retcode
      if( return_value != -1 ):
         for t in range( 0, lastTest ):
            test = IndividuateFields( t, action )
            if ( len(test) > 0 ):
               if ( i < 9 ):
                  info = "  " + str(i+1) + " tst: " + test
                  InfoLine( logfile, info )
               elif ( i < 99 ):
                  info = " " + str(i+1) + " tst: " + test
                  InfoLine( logfile, info )
               else:
                  info = str(i+1) + " tst: " + test
                  InfoLine( logfile, info )
            test_result = EvaluateTest( test, retcode, retoutput, reterr, logfile )
            if ( test_result != 0 ):
               return_value =  1
            if ( test_result == -1):
               return_value = -1
               break
   return( return_value )


def ParseLine( line ):
   separator = 0
   conditions = []
   actions = []
   commands = line.split(",",2)
   for item in commands:
      item = item.strip()
      if separator == 0:
         if ( len(item) == 0 ):
            conditions.append( item )
         elif item[0] == "#":
             separator = 1
         else:
             conditions.append( item )
      else:
            actions.append( item )
   return( conditions, actions )


def BifurcateRules( ruleset ):
   conditions = []
   actions = []
   toggle = 0
   for item in ruleset:
      if ( toggle == 0 ):
         conditions.append( item )
         toggle = 1
      else:
         actions.append( item )
         toggle = 0
   return( conditions, actions )


def IndividuateLines( index, conditions, actions ):
   return( conditions[index], actions[index] )


def IndividuateFields( index, field ):
   return( (field[index]).strip() )


def PerformAction( command, logfile ):
   # TOUCH <file>
   # WAIT ON FILE PRESENT <file>
   # WAIT ON FILE ABSENT <file>
   # WAIT <seconds>
   # CAPTURE <sourcefile> <number> TO <targetfile>
   # START <command>    (runs command asynchronously, doesn't capture output or return code)
   # <command>          (runs command synchronously, waits for output and return code)
   err = ""
   output = ""
   retval = 0
   position = 0
   synchronous = 1
   time_waiting = 0
   quote_location = []
   apostrophe_location = []
   end = len( command )
   while( position < end ):
      position = ( command.find( '"', position, end ) )
      if( position != -1 ):
         quote_location.append( position )
         position += 1
      else:
         position = end
   last_quote = len( quote_location ) - 1
   position = 0
   while( position < end ):
      position = ( command.find( "'", position, end ) )
      if( position != -1 ):
         apostrophe_location.append( position )
         position += 1
      else:
         position = end
   last_apostrophe = len( apostrophe_location ) - 1
   if( last_quote == 0 ):
      info = "==> ERROR: Invalid syntax.  Single quote found.  Please examine ruleset."
      InfoLine( logfile, info )
      retval = -1
   if( last_apostrophe == 0 ):
      if( last_quote > 0 ):
         if( (int(apostrophe_location[0]) < int(quote_location[0])) or (int(apostrophe_location[0]) > int(quote_location[-1])) ):
            info = "==> ERROR: Invalid syntax.  Apostrophe found outside quotes.  Please examine ruleset."
            InfoLine( logfile, info )
            retval = -1
      else:
            info = "==> ERROR: Invalid syntax.  Single apostrophe found.  Please examine ruleset."
            InfoLine( logfile, info )
            retval = -1
   if( retval == 0 ):
      arg = shlex.split( command )
      argc = len(arg)
      if( argc == 0 ):
         retval = 0
      elif( str(arg[0]).isspace() ):
         retval = 0
      elif( str(arg[0]).lower() == "touch" ):
         if( argc != 2 ):
            info = "==> ERROR: Invalid syntax for [TOUCH] rule, wrong number of arguments.  Please examine ruleset."
            InfoLine( logfile, info )
            retval = -1
         else:
            retval = TouchFile( str(arg[1]) )
      elif( str(arg[0]).lower() == "wait" ):
         if( argc == 2 ):
            if( arg[1].isdigit() ):
               seconds = int(arg[1])
               time.sleep( seconds )
            else:
               info = "==> ERROR: Invalid syntax for [WAIT] rule, invalid number of seconds.  Please examine ruleset."
               InfoLine( logfile, info )
               retval = -1
         elif( argc == 5 ):
            if( (str(arg[1]).lower() == "on") and (str(arg[2]).lower() == "file") ):
               lockfile = str(arg[4])
               if( str(arg[3]).lower() == "absent" ):
                  while( os.path.isfile(lockfile) ):
                     time.sleep( POLLING_INTERVAL )
                     time_waiting +=1
                     if( time_waiting > EXCESSIVE_TIMEOUT ):
                        info = "==> ERROR: Timeout occured waiting for [" + lockfile + "] file to become absent."
                        InfoLine( logfile, info )
                        retval = -1
                        break
               elif( str(arg[3]).lower() == "present" ):
                  while( not (os.path.isfile(lockfile)) ):
                     time.sleep( POLLING_INTERVAL )
                     time_waiting +=1
                     if( time_waiting > EXCESSIVE_TIMEOUT ):
                        info = "==> ERROR: Timeout occured waiting for [" + lockfile + "] file to become present."
                        InfoLine( logfile, info )
                        retval = -1
                        break
               else:
                  info = "==> ERROR: Invalid syntax for [WAIT ON FILE] rule, must be ABSENT or PRESENT.  Please examine ruleset."
                  InfoLine( logfile, info )
                  retval = -1
            else:
                  info = "==> ERROR: Invalid syntax for [WAIT] rule, must be ON FILE ABSENT or PRESENT.  Please examine ruleset."
                  InfoLine( logfile, info )
                  retval = -1
         else:
            info = "==> ERROR: Invalid syntax for [WAIT] rule, wrong number of arguments.  Please examine ruleset."
            InfoLine( logfile, info )
            retval = -1
      elif( str(arg[0]).lower() == "capture" ):
         if( argc != 5 ):
            info = "==> ERROR: Invalid syntax for [CAPTURE] rule, wrong number of arguments.  Please examine ruleset."
            InfoLine( logfile, info )
            retval = -1
         elif( str(arg[3]).lower() != "to" ):
            info = "==> ERROR: Invalid syntax for [CAPTURE] rule, missing [TO] keyword.  Please examine ruleset."
            InfoLine( logfile, info )
            retval = -1
         else:
            retval = Capture( arg[1], int(arg[2]), arg[4], logfile )
      else:
         if( str(arg[0]).lower() == "start" ):
            if( argc < 2 ):
               info = "==> ERROR: Invalid syntax for [START] rule, nothing to run.  Please examine ruleset."
               InfoLine( logfile, info )
               retval = -1
            else:
               del arg[0]
               synchronous = 0
         try:
            if( sys.version_info[0] < 3 ):
               proc = subprocess.Popen( command, shell=True,
                                        universal_newlines=True,
                                        stdout=subprocess.PIPE,
                                        stderr=subprocess.PIPE )
            else:
               proc = subprocess.Popen( arg, shell=True,
                                        universal_newlines=True,
                                        stdout=subprocess.PIPE,
                                        stderr=subprocess.PIPE )
            if( synchronous ):
               (output, err) = proc.communicate()
               retval = proc.wait()
         except OSError:
            info = "==> ERROR: Invalid command.  Syntax error or invalid program name.  Please examine ruleset."
            InfoLine( logfile, info )
            retval = -1
   return( retval, output, err )


def EvaluateTest( test, proc_rc, proc_stdout, proc_stderr, logfile ):
   # PASS/FAIL/WARN/EXIT
   # if equal / if not equal <value>         (checks program return code)
   # if has bits <value>
   # if error                                (same as if not equal 0)
   # if exists <filename>
   # if missing  <filename>
   # if contains <pattern> <filename>
   # if tail <filename> <number> contains <pattern>
   # if stdout [offset] equal <pattern>      (looks at stdout string like args, offset optional, default 0)
   # if stdout [offset] not equal <pattern>
   # if stdout [offset] has bits <value>
   # if stdout contains <pattern>            (examines entire stdout string)
   # if stderr [offset] equal <pattern>      (looks at stderr string like args, offset optional, default 0)
   # if stderr [offset] not equal <pattern>
   # if stderr [offset] has bits <value>
   # if stderr contains <pattern>            (examines entire stderr string)
   # if output [offset] equal <pattern>      (looks at both stdout and stderr)
   # if output [offset] not equal <pattern>
   # if output [offset] has bits <value>
   # if output contains <pattern>
   retval = 0
   condition = ""
   verb = "PASS"
   noun = ""
   target = ""
   position = 0
   offset = -1
   intCfgValue = -1
   intNoun = -1
   quote_location = []
   apostrophe_location = []
   end = len( test )
   stdout = proc_stdout.strip()
   stderr = proc_stderr.strip()
   output = stdout + stderr
   while( position < end ):
      position = ( test.find( '"', position, end ) )
      if( position != -1 ):
         quote_location.append( position )
         position += 1
      else:
         position = end
   last_quote = len( quote_location ) - 1
   position = 0
   while( position < end ):
      position = ( test.find( "'", position, end ) )
      if( position != -1 ):
         apostrophe_location.append( position )
         position += 1
      else:
         position = end
   last_apostrophe = len( apostrophe_location ) - 1
   if( last_quote == 0 ):
      info = "==> ERROR: Invalid syntax.  Single quote found.  Please examine ruleset."
      InfoLine( logfile, info )
      retval = -1
   if( last_apostrophe == 0 ):
      if( last_quote > 0 ):
         if( (int(apostrophe_location[0]) < int(quote_location[0])) or (int(apostrophe_location[0]) > int(quote_location[-1])) ):
            info = "==> ERROR: Invalid syntax.  Apostrophe found outside quotes.  Please examine ruleset."
            InfoLine( logfile, info )
            retval = -1
      else:
            info = "==> ERROR: Invalid syntax.  Single apostrophe found.  Please examine ruleset."
            InfoLine( logfile, info )
            retval = -1
   if( retval == 0 ):
      arg = shlex.split( test )
      argc = len( arg )
      if( argc == 0 ):
         retval = 0
      elif( str(arg[0]).isspace() ):
         retval = 0
      else:
         verb = str(arg[0]).upper()
         if( argc > 2 ):
            condition = str(arg[1]) + str(arg[2])
         if( argc > 3 ):
            if( (str(arg[1]).lower() == "if") and
                ( (str(arg[2]).lower() == "stdout") or
                  (str(arg[2]).lower() == "stderr") or
                  (str(arg[2]).lower() == "output") ) and
                  (str(arg[3]).lower() == "contains") ):
                     condition += str(arg[3])
            else:
               noun = str(arg[3])
         if( argc > 4 ):
            if( str(arg[1]).lower() == "if" ):
               if( str(arg[2]).lower() == "not" ):
                  if( str(arg[3]).lower() == "equal" ):
                     condition = str(arg[1]) + str(arg[2]) + str(arg[3])
                     noun = str(arg[4])
               elif( str(arg[2]).lower() == "has" ):
                  if( str(arg[3]).lower() == "bits" ):
                     condition = str(arg[1]) + str(arg[2]) + str(arg[3])
                     noun = str(arg[4])
               elif( (str(arg[2]).lower() == "tail") and (str(arg[5]).lower() == "contains") ):
                  if( argc == 7 ):
                     condition = str(arg[1]) + str(arg[2]) + str(arg[5])
                     target = str(arg[3])
                     offset = str(arg[4])
                     noun = str(arg[6])
               elif( (str(arg[2]).lower() == "stdout") or
                     (str(arg[2]).lower() == "stderr") or
                     (str(arg[2]).lower() == "output") ) :
                     condition = str(arg[1]) + str(arg[2])
                     if( str(arg[3]).lower() == "contains" ):
                        condition += str(arg[3])
                        if( last_quote > 0 ):
                           start = quote_location[0] + 1
                           end = quote_location[last_quote]
                           search_phrase = test[start:end]
                           noun = search_phrase
                        else:
                           noun = str(arg[4])
                           if( argc > 4):
                              i = 5
                              while( i < argc ):
                                 noun += " " + str(arg[i])
                                 i += 1
                     elif( argc == 5 ):
                        if( str(arg[3]).lower() == "equal" ):
                           condition += str(arg[3])
                           noun = str(arg[4])
                     elif( argc == 6 ):
                        if( (str(arg[4]).lower() == "equal") and (str(arg[3]).lower() != "not")  ):
                           offset = int(arg[3])
                           condition += str(arg[4])
                           noun = str(arg[5])
                        elif( (str(arg[3]).lower() == "not") and (str(arg[4]).lower() == "equal") ):
                           condition += str(arg[3]) + str(arg[4])
                           noun = str(arg[5])
                        elif( (str(arg[3]).lower() == "has") and (str(arg[4]).lower() == "bits") ):
                           condition += str(arg[3]) + str(arg[4])
                           noun = str(arg[5])
                     elif( argc == 7 ):
                        if( (str(arg[4]).lower() == "not") and (str(arg[5]).lower() == "equal") ):
                           offset = int(arg[3])
                           condition += str(arg[4]) + str(arg[5])
                           noun = str(arg[6])
                        elif( (str(arg[4]).lower() == "has") and (str(arg[5]).lower() == "bits") ):
                           offset = int(arg[3])
                           condition += str(arg[4]) + str(arg[5])
                           noun = str(arg[6])
         condition = condition.lower()
         if ( condition == "iferror" ):
            if( proc_rc != 0 ):
               test_condition =  0
            else:
               test_condition = -1 
         elif ( condition == "ifexists" ):
            if( os.path.exists( noun ) ):
               test_condition =  0
            else:
               test_condition = -1
         elif ( condition == "ifmissing" ):
            if( os.path.exists( noun ) ):
               test_condition = -1
            else:
               test_condition =  0
         elif ( condition == "ifequal" ):
            if( proc_rc == int(noun) ):
               test_condition =  0
            else:
               test_condition = -1 
         elif ( condition == "ifnotequal" ):
            if( proc_rc != int(noun) ):
               test_condition =  0
            else:
               test_condition = -1
         elif ( condition == "iftailcontains" ):
            tail_string, tail_error = tail(target, offset, logfile)
            if( tail_error ):
               if( verb.upper() == "PASS" ):
                  test_condition = -1
               else:
                  test_condition =  0
            elif( tail_string.find(noun) != -1 ):
               test_condition =  0
            else:
               test_condition = -1
         elif ( condition == "ifhasbits" ):
            try:
               if( (proc_rc & int(noun, 0)) == int(noun, 0) ):
                  test_condition =  0
               else:
                  test_condition = -1
            except ValueError:
               test_condition = 0
               info = "==> ERROR: Invalid non-integer value for [IF HAS BITS] conditional.  Please examine ruleset."
               InfoLine( logfile, info )
               verb = "EXIT"
         elif ( condition == "ifcontains" ):
            test_condition = 0
            if( argc < 5 ):
               info = "==> ERROR: Invalid syntax for [IF CONTAINS] conditional.  Please examine ruleset."
               InfoLine( logfile, info )
               verb = "EXIT"
            else:
               filename = arg[argc-1]
               if( last_quote > 0 ):
                  start = quote_location[0] + 1
                  end = quote_location[last_quote]
                  search_phrase = test[start:end]
                  test_condition = SearchFile( filename, search_phrase, logfile )
                  if ( test_condition == -1 ):
                     verb = "EXIT"
                     test_condition = 0
               else:
                  info = "==> ERROR: Invalid syntax for [IF CONTAINS] conditional.  No quoted search phrase.  Please examine ruleset."
                  InfoLine( logfile, info )
                  verb = "EXIT"
         elif ( condition == "ifstdoutcontains" ):
            test_condition = 0
            if( len(noun) > 0 ):
               if( stdout.find( noun ) == -1 ):
                  test_condition = 1
            else:
               info = "==> ERROR: Invalid syntax for [IF STDOUT CONTAINS] conditional.  No search phrase.  Please examine ruleset."
               InfoLine( logfile, info )
               verb = "EXIT"
         elif ( condition == "ifstderrcontains" ):
            test_condition = 0
            if( len(noun) > 0 ):
               if( stderr.find( noun ) == -1 ):
                  test_condition = 1
            else:
               info = "==> ERROR: Invalid syntax for [IF STDERR CONTAINS] conditional.  No search phrase.  Please examine ruleset."
               InfoLine( logfile, info )
               verb = "EXIT"
         elif ( condition == "ifoutputcontains" ):
            test_condition = 0
            if( len(noun) > 0 ):
               if( output.find( noun ) == -1 ):
                  test_condition = 1
            else:
               info = "==> ERROR: Invalid syntax for [IF OUTPUT CONTAINS] conditional.  No search phrase.  Please examine ruleset."
               InfoLine( logfile, info )
               verb = "EXIT"
         elif ( condition == "ifstdoutequal" ):
            test_condition = 0
            if( offset < 0 ):
               if( len(noun) > 0 ):
                  if( stdout != noun ):
                     test_condition = 1
               else:
                  info = "==> ERROR: Invalid syntax for [IF STDOUT EQUAL] conditional.  No comparison argument.  Please examine ruleset."
                  InfoLine( logfile, info )
                  verb = "EXIT"
            else:
               segment = stdout.split()
               if( len(segment) > offset ):
                  if( segment[offset] != noun ):
                     test_condition = 1
               else:
                  info = "==> ERROR: Invalid offset in [IF STDOUT " + str(offset) + " EQUAL] conditional.  Index exceeds the (" + str(len(segment)) + ") arguments returned."
                  InfoLine( logfile, info )
                  verb = "EXIT"
         elif ( condition == "ifstderrequal" ):
            test_condition = 0
            if( offset < 0 ):
               if( len(noun) > 0 ):
                  if( stderr != noun ):
                     test_condition = 1
               else:
                  info = "==> ERROR: Invalid syntax for [IF STDERR EQUAL] conditional.  No comparison argument.  Please examine ruleset."
                  InfoLine( logfile, info )
                  verb = "EXIT"
            else:
               segment = stderr.split()
               if( len(segment) > offset ):
                  if( segment[offset] != noun ):
                     test_condition = 1
               else:
                  info = "==> ERROR: Invalid offset in [IF STDERR " + str(offset) + " EQUAL] conditional.  Index exceeds the (" + str(len(segment)) + ") arguments returned."
                  InfoLine( logfile, info )
                  verb = "EXIT"
         elif ( condition == "ifoutputequal" ):
            test_condition = 0
            if( offset < 0 ):
               if( len(noun) > 0 ):
                  if( output != noun ):
                     test_condition = 1
               else:
                  info = "==> ERROR: Invalid syntax for [IF OUTPUT EQUAL] conditional.  No comparison argument.  Please examine ruleset."
                  InfoLine( logfile, info )
                  verb = "EXIT"
            else:
               segment = output.split()
               if( len(segment) > offset ):
                  if( segment[offset] != noun ):
                     test_condition = 1
               else:
                  info = "==> ERROR: Invalid offset in [IF OUTPUT " + str(offset) + " EQUAL] conditional.  Index exceeds the (" + str(len(segment)) + ") arguments returned."
                  InfoLine( logfile, info )
                  verb = "EXIT"
         elif ( condition == "ifstdoutnotequal" ):
            test_condition = 0
            if( offset < 0 ):
               if( len(noun) > 0 ):
                  if( stdout == noun ):
                     test_condition = 1
               else:
                  info = "==> ERROR: Invalid syntax for [IF STDOUT NOT EQUAL] conditional.  No comparison argument.  Please examine ruleset."
                  InfoLine( logfile, info )
                  verb = "EXIT"
            else:
               segment = stdout.split()
               if( len(segment) > offset ):
                  if( segment[offset] == noun ):
                     test_condition = 1
               else:
                  info = "==> ERROR: Invalid offset in [IF STDOUT " + str(offset) + " NOT EQUAL] conditional.  Index exceeds the (" + str(len(segment)) + ") arguments returned."
                  InfoLine( logfile, info )
                  verb = "EXIT"
         elif ( condition == "ifstderrnotequal" ):
            test_condition = 0
            if( offset < 0 ):
               if( len(noun) > 0 ):
                  if( stderr == noun ):
                     test_condition = 1
               else:
                  info = "==> ERROR: Invalid syntax for [IF STDERR NOT EQUAL] conditional.  No comparison argument.  Please examine ruleset."
                  InfoLine( logfile, info )
                  verb = "EXIT"
            else:
               segment = stderr.split()
               if( len(segment) > offset ):
                  if( segment[offset] == noun ):
                     test_condition = 1
               else:
                  info = "==> ERROR: Invalid offset in [IF STDERR " + str(offset) + " NOT EQUAL] conditional.  Index exceeds the (" + str(len(segment)) + ") arguments returned."
                  InfoLine( logfile, info )
                  verb = "EXIT"
         elif ( condition == "ifoutputnotequal" ):
            test_condition = 0
            if( offset < 0 ):
               if( len(noun) > 0 ):
                  if( output == noun ):
                     test_condition = 1
               else:
                  info = "==> ERROR: Invalid syntax for [IF OUTPUT NOT EQUAL] conditional.  No comparison argument.  Please examine ruleset."
                  InfoLine( logfile, info )
                  verb = "EXIT"
            else:
               segment = output.split()
               if( len(segment) > offset ):
                  if( segment[offset] == noun ):
                     test_condition = 1
               else:
                  info = "==> ERROR: Invalid offset in [IF OUTPUT " + str(offset) + " NOT EQUAL] conditional.  Index exceeds the (" + str(len(segment)) + ") arguments returned."
                  InfoLine( logfile, info )
                  verb = "EXIT"
         elif ( condition == "ifstdouthasbits" ):
            if( offset < 0 ):
               if( len(noun) > 0 ):
                  try:
                     if( (int(stdout, 0) & int(noun, 0)) == int(noun, 0) ):
                        test_condition =  0
                     else:
                        test_condition = -1
                  except ValueError:
                     test_condition = 0
                     info = "==> ERROR: Invalid non-integer value for [IF STDOUT HAS BITS] conditional.  Please examine ruleset."
                     InfoLine( logfile, info )
                     verb = "EXIT"
               else:
                  test_condition = 0
                  info = "==> ERROR: Invalid syntax for [IF STDOUT HAS BITS] conditional.  No comparison argument.  Please examine ruleset."
                  InfoLine( logfile, info )
                  verb = "EXIT"
            else:
               segment = stdout.split()
               if( len(segment) > offset ):
                  try:
                     if( (int(segment[offset], 0) & int(noun, 0)) == int(noun, 0) ):
                        test_condition =  0
                     else:
                        test_condition = -1
                  except ValueError:
                     test_condition = 0
                     info = "==> ERROR: Invalid non-integer value for [IF STDOUT " + str(offset) + " HAS BITS] conditional.  Please examine ruleset."
                     InfoLine( logfile, info )
                     verb = "EXIT"
               else:
                  test_condition = 0
                  info = "==> ERROR: Invalid offset in [IF STDOUT " + str(offset) + " HAS BITS] conditional.  Index exceeds the (" + str(len(segment)) + ") arguments returned."
                  InfoLine( logfile, info )
                  verb = "EXIT"
         elif ( condition == "ifstderrhasbits" ):
            if( offset < 0 ):
               if( len(noun) > 0 ):
                  try:
                     if( (int(stderr, 0) & int(noun, 0)) == int(noun, 0) ):
                        test_condition =  0
                     else:
                        test_condition = -1
                  except ValueError:
                     test_condition = 0
                     info = "==> ERROR: Invalid non-integer value for [IF STDERR HAS BITS] conditional.  Please examine ruleset."
                     InfoLine( logfile, info )
                     verb = "EXIT"
               else:
                  test_condition = 0
                  info = "==> ERROR: Invalid syntax for [IF STDERR HAS BITS] conditional.  No comparison argument.  Please examine ruleset."
                  InfoLine( logfile, info )
                  verb = "EXIT"
            else:
               segment = stderr.split()
               if( len(segment) > offset ):
                  try:
                     if( (int(segment[offset], 0) & int(noun, 0)) == int(noun, 0) ):
                        test_condition =  0
                     else:
                        test_condition = -1
                  except ValueError:
                     test_condition = 0
                     info = "==> ERROR: Invalid non-integer value for [IF STDERR " + str(offset) + " HAS BITS] conditional.  Please examine ruleset."
                     InfoLine( logfile, info )
                     verb = "EXIT"
               else:
                  test_condition = 0
                  info = "==> ERROR: Invalid offset in [IF STDERR " + str(offset) + " HAS BITS] conditional.  Index exceeds the (" + str(len(segment)) + ") arguments returned."
                  InfoLine( logfile, info )
                  verb = "EXIT"
         elif ( condition == "ifoutputhasbits" ):
            if( offset < 0 ):
               if( len(noun) > 0 ):
                  try:
                     if( (int(output, 0) & int(noun, 0)) == int(noun, 0) ):
                        test_condition =  0
                     else:
                        test_condition = -1
                  except ValueError:
                     test_condition = 0
                     info = "==> ERROR: Invalid non-integer value for [IF OUTPUT HAS BITS] conditional.  Please examine ruleset."
                     InfoLine( logfile, info )
                     verb = "EXIT"
               else:
                  test_condition = 0
                  info = "==> ERROR: Invalid syntax for [IF OUTPUT HAS BITS] conditional.  No comparison argument.  Please examine ruleset."
                  InfoLine( logfile, info )
                  verb = "EXIT"
            else:
               segment = output.split()
               if( len(segment) > offset ):
                  try:
                     if( (int(segment[offset], 0) & int(noun, 0)) == int(noun, 0) ):
                        test_condition =  0
                     else:
                        test_condition = -1
                  except ValueError:
                     test_condition = 0
                     info = "==> ERROR: Invalid non-integer value for [IF OUTPUT " + str(offset) + " HAS BITS] conditional.  Please examine ruleset."
                     InfoLine( logfile, info )
                     verb = "EXIT"
               else:
                  test_condition = 0
                  info = "==> ERROR: Invalid offset in [IF OUTPUT " + str(offset) + " HAS BITS] conditional.  Index exceeds the (" + str(len(segment)) + ") arguments returned."
                  InfoLine( logfile, info )
                  verb = "EXIT"
         else:
            if( argc == 1 ):
               invalid = arg[0]
               info = "==> ERROR: [" + invalid + "] isolated without test conditions.  Please examine ruleset."
               InfoLine( logfile, info )
            else:
               invalid = arg[1]
               for z in range( 2, argc ):
                  invalid = invalid + " " + arg[z]
               info = "==> ERROR: [" + invalid + "] not recognized.  Please examine ruleset."
               InfoLine( logfile, info )
            verb = "EXIT"
            test_condition = 0
 
         localtime = getDateTime()

         if ( verb == "PASS" ):
            if ( test_condition ==  0 ):
               retval =  0
               info = "==> PASS at " + localtime
               InfoLine( logfile, info )
            else:
               retval =  1
               info = "==> FAIL at " + localtime
               InfoLine( logfile, info )
         if ( verb == "FAIL" ):
            if ( test_condition != 0 ):
               retval =  0
               info = "==> PASS at " + localtime
               InfoLine( logfile, info )
            else:
               retval =  1
               info = "==> FAIL at " + localtime
               InfoLine( logfile, info )
         if ( verb == "WARN" ):
            if ( test_condition ==  0 ):
               retval =  0
               info = "==> WARN at " + localtime
               InfoLine( logfile, info )
            else:
               retval =  0
               info = "==> PASS at " + localtime
               InfoLine( logfile, info )
         if ( verb == "EXIT" ):
            if ( test_condition ==  0 ):
               retval = -1
               info = "==> EXIT at " + localtime
               InfoLine( logfile, info )
            else:
               retval =  0
               info = "==> PASS at " + localtime
               InfoLine( logfile, info )
   return( retval )


def Capture( sourcefile, lines_requested, targetfile, logfile ):
   linecount = 0
   delays = 0
   if( os.path.isfile(sourcefile) ):
      try:
         tailfile = open( sourcefile, "r" )
         tailfile.seek( 0, os.SEEK_END )   
         while( linecount < lines_requested ):
            if ( delays < EXCESSIVE_DELAY ):
               marker = tailfile.tell()
               line = tailfile.readline()
               if not line:
                  delays += 1
                  tailfile.close
                  time.sleep( POLLING_INTERVAL )
                  tailfile = open( sourcefile, "r" )
                  tailfile.seek( 0, os.SEEK_END )   
                  tailfile.seek( marker )
               else:
                  delays = 0
                  linecount += 1
                  WriteLine( targetfile, line, logfile )
            else:
               info = "==> ERROR: Timeout occured waiting for activity in [" + sourcefile + "] file."
               InfoLine( logfile, info )
               break
         tailfile.close
      except IOError:
         info = "==> ERROR: [" + sourcefile + "] can't be opened for reading."
         InfoLine( logfile, info )
   else:
      info = "==> ERROR: [" + sourcefile + "] does not exist."
      InfoLine( logfile, info )
   return


def tail( filename, lines_requested, logfile ):
   delays = 0
   section = []
   tail_string = ""
   tail_error = False
   lines_requested = int( lines_requested )
   if( lines_requested > 0 ):
      BUFSIZE = LINE_BUFSIZE * lines_requested
      if( os.path.isfile(filename) ):
         FileLocked = True
         while( FileLocked ):
            if ( delays < EXCESSIVE_DELAY ):
               try:
                  tailfile = open( filename, "rb" )
                  if( tailfile ):
                     FileLocked = False
                     delays = 0
               except:
                  time.sleep( POLLING_INTERVAL )
                  delays += 1
            else:
               tail_error = True
               info = "==> ERROR: Timeout occured waiting for [" + filename + "] file to be unlocked for reading."
               InfoLine( logfile, info )
               break
         if( not FileLocked ):
            block_counter = -1
            while( len(section) < lines_requested ):
               try:
                  tailfile.seek( block_counter * BUFSIZE, os.SEEK_END )
               except IOError:
                  tailfile.seek(0)
                  section = tailfile.readlines()
                  break
               section = tailfile.readlines()
            tailfile.close()
            for line in section[-lines_requested:]:
               tail_string += line.decode(errors='ignore')
         else:
            tail_error = True
            info = "==> ERROR: [" + filename + "] can't be opened for reading."
            InfoLine( logfile, info )
      else:
         tail_error = True
         info = "==> ERROR: [" + filename + "] does not exist."
         InfoLine( logfile, info )
   return( tail_string, tail_error )

