#!/usr/bin/python

'''
   File name: TlogWpp.py
   Author: Wayne Parham
   Date created: 09/01/2019
'''


from __future__ import division
from TlogMetr import *
import string, sys


date=None
details=None
transaction=None
argc = len(sys.argv)
if( (argc < 2) or (argc > 5) ):
   print( "\nUsage: TlogWpp.py tlogfile [date] [transaction] [details]\n" )
   print( "   where:\n")
   print( "   (required) 'tlogfile' is the name of a valid TLOG file," )
   print( "   (optional) 'date' is a 'YYMMDD' formatted date," )
   print( "   (optional) 'transaction' is a four-digit transaction number, and" )
   print( "   (optional) 'details' is a flag that generates detailed output.\n" )
else:
   if( argc > 2 ):
      if( sys.argv[2] == "d" or sys.argv[2] == "-d" or sys.argv[2] == "dump" or
          sys.argv[2] == "display" or sys.argv[2] == "details" or sys.argv[2] == "det" ):
         details = True
      elif( len(sys.argv[2]) == 6 ):
         date = sys.argv[2]
      elif( len(sys.argv[2]) == 4 ):
         transaction = sys.argv[2]
   if( argc > 3):
      if( sys.argv[3] == "d" or sys.argv[3] == "-d" or sys.argv[3] == "dump" or
          sys.argv[3] == "display" or sys.argv[3] == "details" or sys.argv[3] == "det" ):
         details = True
      elif( len(sys.argv[3]) == 6 ):
         date = sys.argv[3]
      elif( len(sys.argv[3]) == 4 ):
         transaction = sys.argv[3]
   if( argc > 4):
      if( sys.argv[4] == "d" or sys.argv[4] == "-d" or sys.argv[4] == "dump" or
          sys.argv[4] == "display" or sys.argv[4] == "details" or sys.argv[4] == "det" ):
         details = True
      elif( len(sys.argv[4]) == 6 ):
         date = sys.argv[4]
      elif( len(sys.argv[4]) == 4 ):
         transaction = sys.argv[4]
   filename = sys.argv[1]
   data = GetDataFromFile( filename )
   try:
      tloglist = ParseTlog( data )
      trans_count, item_count, plan_count, ring_time, tender_time, other_time, wpp_time, max_wpp_time = CarePlanCount( tloglist, date=date, transaction=transaction, details=details )
      if( not transaction ):
         print( "\nTotal transactions \t" + str(trans_count) )
         print( "Total items \t\t" + str(item_count) )
         print( "Total plans \t\t" + str(plan_count) )
         if( ring_time > 3600 ):
            print( "Total ring time \t" + str(round((ring_time / 3600),2)) + " hours" )
         elif( ring_time > 60 ):
            print( "Total ring time \t" + str(round((ring_time / 600),2)) + " minutes" )
         else:
            print( "Total ring time \t" + str(ring_time) + " seconds" )
         if( tender_time > 3600 ):
            print( "Total tender time \t" + str(round((tender_time / 3600),2)) + " hours" )
         elif( tender_time > 60 ):
            print( "Total tender time \t" + str(round((tender_time / 600),2)) + " minutes" )
         else:
            print( "Total tender time \t" + str(tender_time) + " seconds" )
         if( other_time > 3600 ):
            print( "Total 'other' time \t" + str(round((other_time / 3600),2)) + " hours" )
         elif( other_time > 60 ):
            print( "Total 'other' time \t" + str(round((other_time / 600),2)) + " minutes" )
         else:
            print( "Total 'other' time \t" + str(other_time) + " seconds" )
         if( wpp_time > 3600 ):
            print( "Total wpp time \t\t" + str(round((wpp_time / 3600),2)) + " hours" )
         elif( wpp_time > 60 ):
            print( "Total wpp time \t\t" + str(round((wpp_time / 600),2)) + " minutes" )
         else:
            print( "Total wpp time \t\t" + str(round(wpp_time,2)) )
         seconds_per_item = ring_time / item_count
         seconds_per_plan = wpp_time / plan_count
         print( "Avg. seconds per item \t" + str(round(seconds_per_item,2)) )
         print( "Avg. seconds per wpp \t" + str(round(seconds_per_plan,2)) )
         print( "Longest wpp seconds \t" + str(max_wpp_time ) )
   except:
      print( "\nError: Invalid TLOG file.  Please check contents and try again." )

